import datetime
import dateutil.parser
import dateutil.tz.tz
import re

def construct_yaml_timestamp(loader, node):
    loaded_dt = loader.construct_yaml_timestamp(node)
    if type(loaded_dt) is datetime.datetime and loaded_dt.tzinfo is None:
        timezone_match = re.search(
            r'(Z|(?P<sign>[\+-])(?P<h>\d{2}):(?P<m>\d{2}))$',
            loader.construct_yaml_str(node),
            )
        if timezone_match:
            # loader.construct_yaml_timestamp converts to UTC
            loaded_dt = loaded_dt.replace(tzinfo = dateutil.tz.tz.tzutc())
            timezone_attr = timezone_match.groupdict()
            if timezone_attr['h']:
                timezone = dateutil.tz.tz.tzoffset(
                        name = timezone_match.group(0),
                        offset = (int(timezone_attr['h']) * 60 + int(timezone_attr['m'])) * 60
                            * (-1 if timezone_attr['sign'] == '-' else 1),
                        )
                loaded_dt = loaded_dt.astimezone(timezone)
    return loaded_dt

def register_yaml_timestamp_constructor(loader, tag = u'tag:yaml.org,2002:timestamp'):
    loader.add_constructor(tag, construct_yaml_timestamp)

class Duration(object):

    yaml_tag = u'!duration'

    def __init__(self, years = 0):
        self.years = years

    @property
    def years(self):
        return self._years

    @years.setter
    def years(self, years):
        if not type(years) is int:
            raise TypeError('expected int, %r given' % years)
        elif years < 0:
            raise ValueError('number of years must be >= 0, %r given' % years)
        else:
            self._years = years

    @property
    def isoformat(self):
        return 'P%dY' % self.years

    def __eq__(self, other):
        return (type(self) == type(other)
                and self.years == other.years)

    @classmethod
    def from_yaml(cls, loader, node):
        return cls(**loader.construct_mapping(node))

    @classmethod
    def register_yaml_constructor(cls, loader, tag = yaml_tag):
        loader.add_constructor(tag, cls.from_yaml)

    @classmethod
    def to_yaml(cls, dumper, duration, tag = yaml_tag):
        return dumper.represent_mapping(
            tag = tag,
            mapping = {k: v for k, v in {
                'years': duration.years,
                }.items() if v != 0},
            )

    @classmethod
    def register_yaml_representer(cls, dumper):
        dumper.add_representer(cls, cls.to_yaml)

class Period(object):

    yaml_tag = u'!period'

    _timestamp_iso_format = r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(.\d+)?(Z|[-\+]\d{2}:\d{2})?'
    _timeperiod_iso_format = r'(?P<start>%(t)s)\/(?P<end>%(t)s)' % {'t': _timestamp_iso_format}

    def __init__(self, start = None, end = None, isoformat = None):
        self._start = None
        self._end = None
        if (start or end) and isoformat:
            raise AttributeError('when providing isoformat no other parameters may be specified')
        elif isoformat:
            self.isoformat = isoformat
        else:
            self.start = start
            self.end = end

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start):
        if not (start is None or type(start) is datetime.datetime):
            raise TypeError()
        self._start = start

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end):
        if not (end is None or type(end) is datetime.datetime):
            raise TypeError()
        self._end = end

    @property
    def isoformat(self):
        if self.start is None or self.end is None:
            raise ValueError('both start and end must be set')
        return '%s/%s' % (
                self.start.isoformat().replace('+00:00', 'Z'),
                self.end.isoformat().replace('+00:00', 'Z'),
                )

    @isoformat.setter
    def isoformat(self, text):
        match = re.search('^%s$' % self.__class__._timeperiod_iso_format, text)
        if not match:
            raise ValueError(
                    "given string '%s' does not match the supported pattern '%s'"
                     % (text, self.__class__._timeperiod_iso_format)
                     )
        attr = match.groupdict()
        self.start = dateutil.parser.parse(attr['start'])
        self.end = dateutil.parser.parse(attr['end'])

    def __eq__(self, other):
        return (type(self) == type(other)
            and self.start == other.start
            and self.end == other.end)

    @classmethod
    def from_yaml(cls, loader, node):
        return cls(**loader.construct_mapping(node))

    @classmethod
    def to_yaml(cls, dumper, period):
        return dumper.represent_mapping(
            tag = cls.yaml_tag,
            mapping = {
                'start': period.start,
                'end': period.end,
                },
            # represent datetime objects with !timestamp tag
            flow_style = False,
            )

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, ', '.join([
            'start = %r' % self.start,
            'end = %r' % self.end,
            ]))

    @classmethod
    def register_yaml_constructor(cls, loader, tag = yaml_tag):
        register_yaml_timestamp_constructor(loader)
        loader.add_constructor(tag, cls.from_yaml)

    @classmethod
    def register_yaml_representer(cls, dumper):
        dumper.add_representer(cls, cls.to_yaml)



import copy
try:
    import yaml
    import yaml.nodes
except ImportError:
    yaml = None


class UnitMismatchError(ValueError):
    pass


class Figure(object):

    yaml_tag = u"!figure"

    def __init__(self, value=None, unit=None):
        self.value = value
        self.unit = unit

    def get_value(self):
        return self._value

    def set_value(self, value):
        self._value = copy.deepcopy(value)

    """ use property() instead of decorator to enable overriding """
    value = property(get_value, set_value)

    def get_unit(self):
        return self._unit

    def set_unit(self, unit):
        self._unit = copy.deepcopy(unit)

    """ use property() instead of decorator to enable overriding """
    unit = property(get_unit, set_unit)

    def __repr__(self):
        return '{}(value = {!r}, unit = {})'.format(type(self).__name__, self.value, self.unit)

    def __str__(self):
        if self.value is None and self.unit is None:
            return '?'
        elif self.unit is None:
            return '{}'.format(self.value)
        elif self.value is None:
            return '? {}'.format(self.unit)
        else:
            return '{} {}'.format(self.value, self.unit)

    @classmethod
    def from_yaml(cls, loader, node):
        if isinstance(node, yaml.nodes.ScalarNode):
            seg = loader.construct_scalar(node).split(' ')
            if seg[0] == '?':
                value = None
            else:
                try:
                    value = int(seg[0])
                except ValueError:
                    value = float(seg[0])
            return cls(
                value=value,
                unit=' '.join(seg[1:]) if len(seg) > 1 else None,
            )
        else:
            return cls(**loader.construct_mapping(node, deep=True))

    @classmethod
    def register_yaml_constructor(cls, loader, tag=None):
        loader.add_constructor(
            cls.yaml_tag if tag is None else tag,
            cls.from_yaml,
        )

    @classmethod
    def to_yaml(cls, dumper, figure, tag=None):
        tag = tag or cls.yaml_tag
        if figure.value is None or type(figure.value) in [int, float]:
            if figure.value is None:
                value_text = u'?'
            else:
                value_text = u'{}'.format(figure.value)
            if figure.unit is None:
                figure_text = value_text
            else:
                figure_text = u'{} {}'.format(value_text, figure.unit)
            return dumper.represent_scalar(tag=tag, value=figure_text)
        else:
            return dumper.represent_mapping(
                tag=tag,
                mapping={'value': figure.value, 'unit': figure.unit},
            )

    @classmethod
    def register_yaml_representer(cls, dumper, tag=None):
        tag = tag or cls.yaml_tag
        dumper.add_representer(cls, lambda d, f: cls.to_yaml(d, f, tag=tag))

    def __eq__(self, other):
        return isinstance(self, type(other)) and vars(self) == vars(other)

    def __ne__(self, other):
        return not (self == other)

    def __add__(self, other):
        if not isinstance(self, type(other)):
            raise NotImplementedError('{!r} + {!r}'.format(self, other))
        assert not self.value is None
        assert not other.value is None
        if self.unit != other.unit:
            raise UnitMismatchError('{} + {}'.format(self, other))
        else:
            return type(self)(value=self.value + other.value, unit=self.unit)

    def __radd__(self, other):
        """ enables use of sum() """
        if isinstance(other, int) and other == 0:
            return copy.deepcopy(self)
        else:
            raise NotImplementedError('{!r} + {!r}'.format(other, self))

    def __sub__(self, other):
        if not isinstance(self, type(other)):
            raise NotImplementedError('{!r} - {!r}'.format(self, other))
        assert not self.value is None
        assert not other.value is None
        if self.unit != other.unit:
            raise UnitMismatchError('{} - {}'.format(self, other))
        else:
            return type(self)(value=self.value - other.value, unit=self.unit)

    def __mul__(self, factor):
        if isinstance(factor, Figure):
            assert not self.value is None
            assert not factor.value is None
            assert factor.unit is None
            return type(self)(value=self.value * factor.value, unit=self.unit)
        else:
            return self * Figure(value=factor, unit=None)

    """
    $ python2
    >>> 3/2
    1
    $ python3
    >>> 3/2
    1.5
    >>> 4/2
    2.0
    """

    def __truediv__(self, divisor):
        if isinstance(divisor, Figure):
            assert not self.value is None
            assert not divisor.value is None
            if isinstance(self.value, int):
                value = float(self.value) / divisor.value
            else:
                value = self.value / divisor.value
            if self.unit == divisor.unit:
                return Figure(value=value, unit=None)
            elif divisor.unit is None:
                return type(self)(value=value, unit=self.unit)
            else:
                raise NotImplementedError('{!r} / {!r}'.format(self, divisor))
        else:
            return self / Figure(value=divisor, unit=None)

    def __div__(self, divisor):
        return self.__truediv__(divisor)

    def __round__(self, *params, **kwargs):
        return type(self)(
            value=round(self.value, *params, **kwargs),
            unit=self.unit,
        )



# http://misc.flogisoft.com/bash/tip_colors_and_formatting

class Formatting:
    bold = '\033[1m'
    dim = '\033[2m'
    reset_bold = '\033[21m'
    reset_dim = '\033[22m'
    reset = ''.join([reset_bold, reset_dim])

class TextColor:
    default = '\033[39m'
    red = '\033[31m'
    green = '\033[32m'
    yellow = '\033[33m'
    blue = '\033[34m'
    magenta = '\033[35m'



import contextlib
import difflib
import ioex.shell
import locale
import readline
import threading
try:
    import yaml
except ImportError:
    yaml = None


class UnsupportedLocaleSettingError(locale.Error):
    pass

locale_lock = threading.Lock()


@contextlib.contextmanager
def setlocale(temporary_locale):
    with locale_lock:
        primary_locale = locale.setlocale(locale.LC_ALL)
        try:
            try:
                yield locale.setlocale(locale.LC_ALL, temporary_locale)
            except locale.Error as ex:
                if str(ex) == 'unsupported locale setting':
                    raise UnsupportedLocaleSettingError(temporary_locale)
                else:
                    raise ex
        finally:
            locale.setlocale(locale.LC_ALL, primary_locale)


def raw_input_with_default(prompt, default):
    def pre_input_hook():
        readline.insert_text(default)
        readline.redisplay()
    readline.set_pre_input_hook(pre_input_hook)
    try:
        return raw_input(prompt)
    finally:
        readline.set_pre_input_hook(None)


def int_input_with_default(prompt, default):
    if default:
        default = str(default)
    else:
        default = ''
    s = raw_input_with_default(prompt, default).strip()
    if s:
        return int(s)
    else:
        return None


def yaml_represent_unicode_as_str(dumper, unicode_string):
    return dumper.represent_scalar(u'tag:yaml.org,2002:str', unicode_string)


def register_yaml_unicode_as_str_representer(dumper):
    try:
        dumper.add_representer(unicode, yaml_represent_unicode_as_str)
    except NameError:  # python3
        pass


def yaml_construct_str_as_unicode(loader, node):
    string = loader.construct_scalar(node)
    try:
        return string if isinstance(string, unicode) else string.decode('utf-8')
    except NameError:  # python3
        return string


def register_yaml_str_as_unicode_constructor(loader):
    loader.add_constructor(
        u'tag:yaml.org,2002:str',
        yaml_construct_str_as_unicode,
    )


yaml_diff_colors = {
    ' ': ioex.shell.TextColor.default,
    '+': ioex.shell.TextColor.green,
    '-': ioex.shell.TextColor.red,
    '?': ioex.shell.TextColor.yellow,
}


def yaml_diff(a, b, dumper=None, colors=False):
    if dumper is None:
        class DiffDumper(yaml.Dumper):
            pass
        register_yaml_unicode_as_str_representer(DiffDumper)
        dumper = DiffDumper

    def to_yaml(data):
        return yaml.dump(
            data,
            Dumper=dumper,
            default_flow_style=False,
            allow_unicode=True,
        )
    diff_lines = difflib.ndiff(
        to_yaml(a).splitlines(True),
        to_yaml(b).splitlines(True),
    )
    if colors:
        diff_lines = [u'%s%s%s' % (yaml_diff_colors[l[0]], l, ioex.shell.TextColor.default)
                      for l in diff_lines]
    return u''.join(diff_lines)


class AutoDict(dict):

    def __missing__(self, key):
        value = self[key] = type(self)()
        return value

    @classmethod
    def to_yaml(cls, dumper, collection):
        return dumper.represent_dict(collection)

    @classmethod
    def register_yaml_representer(cls, dumper):
        dumper.add_representer(cls, cls.to_yaml)


def dict_collapse(dict_src, key_repl):
    def single_not_none(sub_dict):
        not_none = [v for v in sub_dict.values() if v is not None]
        if len(not_none) == 0:
            return None
        elif len(not_none) == 1:
            return not_none[0]
        else:
            raise ValueError(sub_dict)

    sub_dicts = AutoDict()
    for key in dict_src:
        sub_dicts[key_repl(key)][key] = dict_src[key]
    return {key: single_not_none(sub) for key, sub in sub_dicts.items()}



from setuptools import setup

import glob

setup(
    name = 'ioex',
    packages = ['ioex'],
    version = '0.13.0',
    description = 'extension for python\'s build-in input / output interface',
    author = 'Fabian Peter Hammerle',
    author_email = 'fabian.hammerle@gmail.com',
    url = 'https://github.com/fphammerle/ioex',
    download_url = 'https://github.com/fphammerle/ioex/tarball/0.13.0',
    keywords = [],
    classifiers = [],
    scripts = glob.glob('scripts/*'),
    install_requires = ['python-dateutil'],
    tests_require = ['pytest', 'pytz'],
    )



import re


def rename_groups(pattern, name_repl):
    def pattern_repl(match):
        return '(?P<{}>'.format(name_repl(match.group(1)))
    return re.sub(
        r'\(\?P<(.*?)>',
        pattern_repl,
        pattern,
    )



import os
import sys
import locale

def tty_wrapper(function, *args, **kwargs):
    stdin_fd = os.dup(sys.stdin.fileno())
    stdout_fd = os.dup(sys.stdout.fileno())
    # set stdin to tty
    tty_in = open('/dev/tty', 'r')
    os.dup2(tty_in.fileno(), sys.stdin.fileno())
    # set stdout to tty
    tty_out = open('/dev/tty', 'w')
    os.dup2(tty_out.fileno(), sys.stdout.fileno())
    # os.system('ls -l /proc/self/fd/')
    locale.setlocale(locale.LC_ALL, '')
    import curses
    result = curses.wrapper(function, *args, **kwargs)
    # reset stdin and stdout
    os.dup2(stdin_fd, sys.stdin.fileno())
    os.dup2(stdout_fd, sys.stdout.fileno())
    return result
